from .service import Service, ConfigurableService
from .config import parse_address
from .util import waitany

from . import _meta


__version__ = _meta.version
__version_info__ = _meta.version_info


__all__ = [
    'ConfigurableService',
    'Service',
    'parse_address',
    'waitany',
]



import functools
import sys

import gevent
import logbook
import pyee

from . import util, config as biloba_config


class Service(pyee.EventEmitter):
    """
    An asynchronous primitive that will maintain a pool of spawned greenlets
    and watch any child services (which are just objects the same as this).

    The service will wait until all children greenlets have completed.

    There is a distinction between arbitrary greenlets and 'service' greenlets.
    A service greenlet is meant to run forever, if for some reason it dies
    early then all other services are torn down as this parent service is done.

    :ivar started: Whether this service has started.
    :type started: Boolean.
    :ivar services: A list of child service objects that this service is
        watching.
    :ivar spawned_greenlets: A list of greenlets that is being watched by this
        service.
    """

    def __init__(self):
        super(Service, self).__init__()

        self.started = False
        self.services = []
        self.spawned_greenlets = []

    def __del__(self):
        try:
            self.stop()
        except:
            pass

    @util.cachedproperty
    def logger(self):
        return self.get_logger()

    # set to specify the logger name (before the first access)
    logger_name = None

    def get_logger(self):
        return logbook.Logger(self.logger_name or self.__class__.__name__)

    def do_start(self):
        """
        Called when this service is starting but before it is actually running.

        Create any child services or spawn greenlets here.
        """

    def do_stop(self):
        """
        Called when this service is stopping.

        Perform any necessary cleanup.
        """

    def start(self):
        """
        Called to start this service and any child services that may be
        registered.

        Any custom code should generally go in to `do_start`.
        """
        if self.started:
            return

        self.do_start()

        for service in self.services:
            service.start()

        self.spawn(self.watch_services)

        self.started = True

        self.emit('start')

    def stop(self):
        """
        Called to stop this service if it is running. All child services are
        `stopped` and any greenlets this service may have spawned are killed.

        Any custom code should generally go in to `do_stop`.
        """
        if not self.started:
            return

        try:
            for service in self.services:
                service.stop()

            for g in self.spawned_greenlets:
                if not g.dead:
                    g.kill()

            self.services = []
            self.spawned_greenlets = []

            self.do_stop()
        finally:
            self.started = False
            self.emit('stop')

    def join(self):
        """
        Called to block the current greenlet to wait for this service to finish
        """
        # start is idempotent
        self.start()

        # all `spawned` greenlets are `link`ed to `remove_greenlet` which means
        # that when `joinall` returns all of the greenlets that were passed in
        # will have been removed from the `spawned_greenlets` list - however,
        # those greenlets may have spawned more greenlets.
        while self.spawned_greenlets:
            gevent.joinall(self.spawned_greenlets)

        self.stop()

    def add_service(self, *services):
        """
        Add a service to this parent service object. A child service must only
        have one parent
        """
        self.services.extend(services)

        if not self.started:
            return

        for service in services:
            self.spawn(service.join)

    def remove_greenlet(self, g):
        try:
            self.spawned_greenlets.remove(g)
        except ValueError:
            pass

    def spawn(self, func, *args, **kwargs):
        """
        Spawns a greenlet that is linked to this service and will be killed if
        the service stops.

        :param func: The callable to execute in a new greenlet context.
        :param args: The args to pass to the callable.
        :param kwargs: The kwargs to pass to the callable.
        :returns: The spawned greenlet thread.
        """
        def log_exc(func):
            """
            Log any exceptions as they happen from the function being spawned.
            """
            @functools.wraps(func)
            def wrapper(*args, **kwargs):  # pylint: disable=C0111
                try:
                    return func(*args, **kwargs)
                except:
                    self.emit('error', *sys.exc_info())

            return wrapper

        thread = gevent.spawn(log_exc(func), *args, **kwargs)

        self.spawned_greenlets.append(thread)

        thread.rawlink(self.remove_greenlet)

        return thread

    def watch_services(self):
        # `self.spawn` is not used because we don't use want to kill these
        # immediately, see below
        greenlets = [gevent.spawn(s.join) for s in self.services]

        if not greenlets:
            return

        # a service greenlet is never supposed to exit, so if one does then
        # the parent service has ended and any calls to `join` must exit.
        ret = util.waitany(greenlets)

        greenlets.remove(ret)

        for thread in greenlets:
            thread.kill()

        if ret.exception:
            raise ret.exception.__class__, ret.exception, None


class ConfigurableService(Service):
    """
    A service that takes a config dict
    """

    def __init__(self, config):
        """
        :param config: Provide a dict like interface
        """
        super(ConfigurableService, self).__init__()

        config = self.apply_default_config(config or {})

        self.config = biloba_config.Config(config)

    def get_config_defaults(self):
        return {}

    def apply_default_config(self, config):
        defaults = self.get_config_defaults()

        for key, value in defaults.items():
            config.setdefault(key, value)

        return config



from gevent import event


class cachedproperty(object):
    """
    A decorator that converts a method in to a property and caches the value
    returned by that method for fast retrieval.
    """

    def __init__(self, func):
        self.func = func

    def __get__(self, obj, *args, **kwargs):
        if obj is None:
            return self

        value = self.func(obj)

        setattr(obj, self.func.__name__, value)

        return value


def waitany(greenlets, timeout=None, result_class=event.AsyncResult):
    """
    Given a list of greenlets, wait for the first one to return a result.

    Note that only the greenlet is returned, not the value (or the exception).

    :param greenlets: A list of greenlets.
    :param timeout: The maximum amount of time to wait before raising
        `gevent.Timeout`. A timeout of `None` means to wait potentially
        forever.
    :param result_class: Advanced usage and tests.
    :return: The greenlet that first returned a result.
    """
    result = result_class()
    update = result.set

    try:
        for thread in greenlets:
            # start is idempotent :)
            thread.start()

            if not thread.ready():
                thread.rawlink(update)

                continue

            # this greenlet contains a value already
            return thread

        return result.get(timeout=timeout)
    finally:
        for thread in greenlets:
            thread.unlink(update)



"""
Tests for `biloba.service`.
"""

import unittest
import mock

import gevent

from biloba import service


def make_service():
    return service.Service()


class ServiceTestCase(unittest.TestCase):
    """
    Tests for `service.Service`
    """

    def test_create(self):
        """
        Ensure basic attributes when initialising the service
        """
        my_service = make_service()

        self.assertFalse(my_service.started)
        self.assertEqual(my_service.services, [])
        self.assertEqual(my_service.spawned_greenlets, [])

    def test_delete(self):
        """
        Ensure stop is called when deleting the service.
        """
        my_service = make_service()

        with mock.patch.object(my_service, 'stop') as mock_stop:
            my_service.__del__()

            mock_stop.assert_called_once_with()

    def test_delete_error(self):
        """
        If `stop` causes an error while `__del__` is being called, swallow the
        exception.
        """
        my_service = make_service()

        with mock.patch.object(my_service, 'stop') as mock_stop:
            mock_stop.side_effect = RuntimeError

            my_service.__del__()

    def test_get_logger(self):
        """
        A service should provide a logger as an attribute.
        """
        import logbook

        my_service = make_service()

        logger = my_service.logger

        self.assertIsInstance(logger, logbook.Logger)
        self.assertEqual(logger.name, 'Service')

    def test_get_logger_custom_name(self):
        """
        A service logger should be able to be named.
        """
        import logbook

        my_service = make_service()

        my_service.logger_name = 'my_log_name'

        logger = my_service.logger

        self.assertIsInstance(logger, logbook.Logger)
        self.assertEqual(logger.name, 'my_log_name')

    def test_start(self):
        """
        Ensure that starting a service works correctly.
        """
        my_service = make_service()

        self.event_fired = False

        def on_start():
            self.event_fired = True

        my_service.on('start', on_start)

        self.assertFalse(my_service.started)
        my_service.start()
        self.assertTrue(my_service.started)
        self.assertTrue(self.event_fired)

        # now test calling `start` again. It should be idempotent
        self.event_fired = False

        my_service.start()

        self.assertFalse(self.event_fired)

    def test_do_start(self):
        """
        Ensure that do_start is called during `start`.
        """
        my_service = make_service()

        with mock.patch.object(my_service, 'do_start') as mock_do_start:
            my_service.start()

            mock_do_start.assert_called_one_with()

            # ensure that `do_start` is called only once.
            my_service.start()

            mock_do_start.assert_called_one_with()

    def test_start_services(self):
        """
        Any greenlets in `service.services` must be `start`ed.
        """
        my_service = make_service()
        mock_greenlet = mock.Mock()

        my_service.services = [mock_greenlet]

        my_service.start()

        mock_greenlet.start.assert_called_once_with()

    def test_stop_not_started(self):
        """
        Can only `stop` a service when it has been started.
        """
        my_service = make_service()
        self.event_fired = False

        def on_stop():
            self.event_fired = True

        my_service.on('stop', on_stop)

        my_service.stop()

        self.assertFalse(self.event_fired)

    def test_stop(self):
        """
        Ensure that `stop`ping a service works correctly.
        """
        my_service = make_service()

        my_service.start()

        self.event_fired = False

        def on_stop():
            self.event_fired = True

        my_service.on('stop', on_stop)

        self.assertTrue(my_service.started)
        my_service.stop()
        self.assertFalse(my_service.started)
        self.assertTrue(self.event_fired)

        # now test calling `stop` again. It should be idempotent
        self.event_fired = False

        my_service.stop()

        self.assertFalse(self.event_fired)

    def test_do_stop(self):
        """
        Ensure that `do_stop` is called during `stop`.
        """
        my_service = make_service()

        my_service.start()

        with mock.patch.object(my_service, 'do_stop') as mock_do_stop:
            my_service.stop()

            mock_do_stop.assert_called_one_with()

            # ensure that `do_stop` is called only once.
            my_service.stop()

            mock_do_stop.assert_called_one_with()

    def test_stop_services(self):
        """
        Any greenlets in `service.services` must be `stop`ped.
        """
        my_service = make_service()
        my_service.start()

        mock_greenlet = mock.Mock()

        my_service.services = [mock_greenlet]

        my_service.stop()

        self.assertEqual(my_service.services, [])

        mock_greenlet.stop.assert_called_once_with()

    @mock.patch.object(service.Service, 'start')
    @mock.patch.object(service.Service, 'stop')
    def test_join(self, mock_stop, mock_start):
        """
        Ensure that `join` works as one would expect.
        """
        my_service = make_service()

        my_service.spawn(lambda: 'foobar')

        my_service.join()

        mock_start.assert_called_once_with()
        mock_stop.assert_called_with()

    def test_spawn(self):
        """
        Spawning a greenlet via the service allows for special things.
        """
        my_service = make_service()

        thread = my_service.spawn(lambda: [1, 2, 3])

        self.assertIn(thread, my_service.spawned_greenlets)

        # force a couple of context switches which allows the greenlet to be
        # cleaned up
        gevent.sleep(0.0)
        gevent.sleep(0.0)

        self.assertEqual(my_service.spawned_greenlets, [])

    def test_spawn_error(self):
        """
        Spawning a greenlet that raises an exception must emit the error event
        via the service.
        """
        my_service = make_service()

        self.executed = False

        def trap_error(exc_type, exc_value, exc_traceback):
            self.assertIsInstance(exc_value, RuntimeError)
            self.executed = True

        def raise_error():
            raise RuntimeError

        my_service.on('error', trap_error)

        my_service.spawn(raise_error)

        gevent.sleep(0.0)

        self.assertTrue(self.executed)

    def test_add_service(self):
        """
        Adding a child service when the parent service is stopped.
        """
        my_service = make_service()
        mock_service = mock.Mock()

        with mock.patch.object(my_service, 'spawn') as mock_spawn:
            my_service.add_service(mock_service)

            self.assertFalse(mock_spawn.called)

        self.assertEqual(my_service.services, [mock_service])

    def test_add_service_start(self):
        """
        Adding a child service when the parent service is started.
        """
        my_service = make_service()
        mock_service = mock.Mock()

        my_service.start()

        with mock.patch.object(my_service, 'spawn') as mock_spawn:
            my_service.add_service(mock_service)

            mock_spawn.assert_called_with(mock_service.join)

        self.assertEqual(my_service.services, [mock_service])

    def test_watch_service(self):
        """
        Ensure basic services do the right thing.
        """
        my_service = make_service()

        my_greenlet = gevent.spawn(lambda: 5)

        my_service.services = [my_greenlet]

        my_service.watch_services()

    def test_watch_service_error(self):
        """
        Error in a service.
        """
        my_service = make_service()
        service = mock.Mock()

        service.join.side_effect = RuntimeError

        my_service.services = [service]

        with self.assertRaises(RuntimeError):
            my_service.watch_services()


class ConfigurableServiceTestCase(unittest.TestCase):
    """
    Tests for `service.ConfigurableService`.
    """

    def test_create(self):
        """
        Create a configurable service.
        """
        from biloba import config

        my_service = service.ConfigurableService(None)

        self.assertIsInstance(my_service.config, config.Config)

    def test_default_config(self):
        """
        Test default config
        """
        class MyService(service.ConfigurableService):
            def get_config_defaults(self):
                return {'foo': 'bar'}

        my_service = MyService(None)

        self.assertEqual(my_service.config['foo'], 'bar')

    def test_override_default_config(self):
        class MyService(service.ConfigurableService):
            def get_config_defaults(self):
                return {'foo': 'bar'}

        my_service = MyService({'foo': 'baz'})

        self.assertEqual(my_service.config['foo'], 'baz')



from setuptools import find_packages, setup

from pip.req import parse_requirements


def get_version():
    import imp

    with open('biloba/_meta.py', 'rb') as fp:
        mod = imp.load_source('_meta', 'biloba', fp)

    return mod.version


def get_requirements(filename):
    reqs = parse_requirements(filename)

    return [str(r.req) for r in reqs]


def get_install_requires():
    return get_requirements('requirements.txt')


def get_test_requires():
    return get_requirements('requirements_dev.txt')


setup_args = dict(
    name='biloba',
    version=get_version(),
    maintainer='Nick Joyce',
    maintainer_email='nick@boxdesign.co.uk',
    packages=find_packages(),
    install_requires=get_install_requires(),
    tests_require=get_test_requires(),
)


if __name__ == '__main__':
    setup(**setup_args)



"""
Tests for `biloba.config`.
"""

import unittest

from biloba import config


class ParseAddressTestCase(unittest.TestCase):
    """
    Tests for `config.parse_address`
    """

    def test_address(self):
        """
        Ensure that is basically works
        """
        host, port = config.parse_address('foo:1234')

        self.assertEqual(host, 'foo')
        self.assertEqual(port, 1234)

    def test_no_port(self):
        host, port = config.parse_address('foo')

        self.assertEqual(host, 'foo')
        self.assertIsNone(port)

    def test_port_is_not_int(self):
        with self.assertRaises(ValueError):
            config.parse_address('foo:bar')


class GetKeyTestCase(unittest.TestCase):
    """
    Tests for `config.get_key`
    """

    config = {
        'foo': {
            'bar': [1, 2, 3],
        }
    }

    def test_missing_key(self):
        """
        Ensure that the default is returned when the key is missing from the
        config
        """
        self.assertIsNone(config.get_key(self.config, 'bar'))
        missing = object()

        ret = config.get_key(self.config, 'bar', missing)

        self.assertIs(ret, missing)

    def test_not_dotted(self):
        """
        A key on the first layer returns the correct value
        """
        self.assertEqual(
            config.get_key(self.config, 'foo'),
            self.config['foo']
        )

    def test_dotted(self):
        """
        dotted notation must be supported.
        """
        self.assertEqual(
            config.get_key(self.config, 'foo.bar'),
            self.config['foo']['bar']
        )


class ConfigTestCase(unittest.TestCase):
    """
    Tests for `config.Config`
    """

    my_config = {
        'http': {
            'address': '127.0.0.1',
            'port': 4000,
        },
        'logger': {
            'address': '${http.address}',
        }
    }

    def test_sanity(self):
        conf = config.Config(self.my_config)

        self.assertEqual(
            conf.get('logger.address'),
            '127.0.0.1'
        )

    def test_expand_list(self):
        my_config = {
            'foo': {
                'a': 'A',
                'b': 'B',
            },
            'bar': ['${foo.a}', 1, '${foo.b}']
        }

        conf = config.Config(my_config)

        self.assertEqual(
            conf['bar'],
            ['A', 1, 'B'],
        )

    def test_expand_dict(self):
        my_config = {
            'foo': {
                'a': 'A',
                'b': 'B',
            },
            'bar': {
                'c': '${foo.a}',
                'd': '${foo.b}'
            },
        }

        conf = config.Config(my_config)

        self.assertEqual(
            conf['bar'],
            dict(c='A', d='B'),
        )

    def test_setdefault(self):
        a = object()
        b = object()

        conf = config.Config()

        conf.setdefault('a', a)
        self.assertIs(conf['a'], a)
        conf.setdefault('a', b)
        self.assertIs(conf['a'], a)
